/*
    drummer.c
     
    Trommler - X11 Drum Machine
    Copyright (C) 1998-2002 Robert Muth <trommler@muth.org>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; version 2 of June 1991.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program in form of the file COPYING; 
    if not, write to the 

    Free Software Foundation, Inc. <http://www.fsf.org>
    59 Temple Place, Suite 330, 
    Boston, MA 02111-1307  USA

    $Id: drummer.c,v 1.6 2002/11/30 16:59:28 muth Exp $
*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "std.h"
#include "drummer.h"
#include "export.h"
#include "version.h"

/* ==================================================== */
/* ==================================================== */

GLOBALVAR TYPE_GLOBAL Global;

/* ==================================================== */
/* ==================================================== */

GLOBALFUN INT32 FindUnusedVoice()
{
    int voice;
    FOREACH_VOICE(voice)
    {
        if( VOICE_LENGTH(voice)==0 )
            return voice;
    }

    fprintf(stderr,"not enough voices\n");
    return 0;
}


/* ==================================================== */
/* ==================================================== */

GLOBALFUN void VoiceSetToDrum(INT32 voice, INT16 *sample, INT32 length, INT32 volume, float pan)
{
    VOICE_SAMPLE(voice) = sample;
    VOICE_LENGTH(voice) = length;
    VOICE_VOLUME(voice) = volume;
    VOICE_PAN(voice)    = pan;
}
/* ==================================================== */

/* ==================================================== */

GLOBALFUN void BeatAssignDrumsToVoices(INT32 patt,INT32 beat)
{
    int drum;
    int voice;
    INT32 accent = 100;
    
    FOREACH_DRUM(drum)
    {
        if( !PATT_DRUM_BEAT(patt,drum,beat) ) continue;
        if (1 == DRUM_MUTE(drum)) continue;
        if( strcmp( DRUM_NAME(drum), "@ACCENT") ) continue;
        accent = accent * DRUM_VOLUME(drum);
        accent = accent / 100;
    }
    
    FOREACH_DRUM(drum)
    {
        if( !PATT_DRUM_BEAT(patt,drum,beat) ) continue;
        if (1 == DRUM_MUTE(drum)) continue;
        if( !strcmp( DRUM_NAME(drum), "@ACCENT") ) continue;

        voice = FindUnusedVoice();
        VoiceSetToDrum(voice,DRUM_SAMPLE(drum),DRUM_LENGTH(drum),DRUM_VOLUME(drum)*accent/100, DRUM_PAN(drum));
#if 0
           fprintf(stderr,"setting voice %d to drum %d\n",voice,drum);
#endif
    }
}

/* ==================================================== */

/* ==================================================== */

GLOBALFUN void Convert32To16Bit(INT32 *src_l, INT32 *src_r, INT16 *dst, INT32 length)
{
    INT32 sample;

    for(sample=0; sample<length; sample++)
    {
      // LEFT Channel
      //
      // Handle clipping...
      if( src_l[sample] >= (1<<15) )
        dst[2*sample] = (1<<15)-1;
      else if( src_l[sample] < -(1<<15) )
        dst[2*sample] = -(1<<15);
      // else just copy the data
      else
        dst[2*sample] = src_l[sample];

      // RIGHT Channel
      if( src_r[sample] >= (1<<15) )
        dst[2*sample+1] = (1<<15)-1;
      else if( src_r[sample] < -(1<<15) )
        dst[2*sample+1] = -(1<<15);
      else
        dst[2*sample+1] = src_r[sample];
    }
}

/* ==================================================== */

/* ==================================================== */

GLOBALFUN void VoicesExtractTo32Bit(INT32 *data_l, INT32 *data_r, INT32 length)
{
    INT32 voice;
    INT32 sample;
    INT32 wave;
    
    for(sample=0;sample < length; sample++)
    {
        data_l[sample] = 0;
        data_r[sample] = 0;
    }
   
   
    FOREACH_VOICE(voice)
    {
        if( VOICE_LENGTH(voice)==0 ) continue;
        if( VOICE_LENGTH(voice) < length )
        {
            for(sample=0;sample < VOICE_LENGTH(voice); sample++) {
                wave = VOICE_VOLUME(voice)*VOICE_SAMPLE(voice)[sample]/100;
                data_l[sample] += wave + wave * -1 * VOICE_PAN(voice);
                data_r[sample] += wave + wave * VOICE_PAN(voice);
            }
            VOICE_SAMPLE(voice) = 0;
            VOICE_LENGTH(voice) = 0;
        }
        else
        {
            for(sample=0;sample < length; sample++) {
                wave = VOICE_VOLUME(voice)*VOICE_SAMPLE(voice)[sample]/100;
                data_l[sample] += wave + wave * -1 * VOICE_PAN(voice);
                data_r[sample] += wave + wave * VOICE_PAN(voice);
            }
            VOICE_SAMPLE(voice) += length;
            VOICE_LENGTH(voice) -= length;
        }
    }
}


/* ==================================================== */

/* ==================================================== */

GLOBALFUN void VoiceClear(INT32 voic)
{
    VOICE_LENGTH(voic) = 0;
    VOICE_SAMPLE(voic) = 0;
}

/* ==================================================== */

/* ==================================================== */

GLOBALFUN void SongClear()
{
    INT32 pos;
    INT32 drum;
    INT32 pattern;
    
    strcpy(SONG_NAME,"No Name");
    strcpy(SONG_PATH,"");
    SONG_LENGTH = 0;
    SONG_UPMS = 60;
    FOREACH_SONG_MEASURE(pos)
        SONG_MEASURE[pos] = 0;

    FOREACH_DRUM(drum)
    {
        DrumClear(drum);
    }

    FOREACH_PATT(pattern)
    {
        PattClear(pattern);
    }
}

/* ==================================================== */

/* ==================================================== */

GLOBALFUN void PattClear(INT32 patt)
{
    INT32 beat;
    INT32 drum;
        
    /* INT32 drum,beat; */
    
    sprintf(PATT_NAME(patt),"Pattern %3d",patt);
    PATT_UNITS(patt) = 4;
    PATT_BEATS_PER_UNIT(patt) = 4;
    FOREACH_DRUM(drum)
        for(beat=0;beat <MAX_BEAT_NUM;beat++)
            PATT_DRUM_BEAT(patt,drum,beat) = 0;
}

/* ==================================================== */

/* ==================================================== */

GLOBALFUN void PattCopy(INT32 dst, INT32 src)
{
    INT32 beat;
    INT32 drum;
        
    /* INT32 drum,beat; */

    strcpy( PATT_NAME(dst),PATT_NAME(src));

    PATT_UNITS(dst) = PATT_UNITS(src);
    PATT_BEATS_PER_UNIT(dst) = PATT_BEATS_PER_UNIT(src);
    
    FOREACH_DRUM(drum)
        for(beat=0;beat <MAX_BEAT_NUM;beat++)
            PATT_DRUM_BEAT(dst,drum,beat) = PATT_DRUM_BEAT(src,drum,beat);
}

/* ==================================================== */
/* ==================================================== */

GLOBALFUN void DrumClear(INT32 drum)
{
    sprintf(DRUM_NAME(drum),"Drum %2d",drum);
    strcpy(DRUM_PATH(drum),"");
    DRUM_SAMPLE(drum) = 0;
    DRUM_KEY(drum) = 0;
    DRUM_LENGTH(drum) = 0;
    DRUM_VOLUME(drum) = 0;
    DRUM_PAN(drum) = 0.0;
    DRUM_MUTE(drum) = 0;
}

/* ==================================================== */
/* ==================================================== */

GLOBALFUN void ClearAll()
{
    int i;
    SongClear();
    FOREACH_DRUM(i)
        DrumClear(i);
    FOREACH_PATT(i)
        PattClear(i);
}

/* ==================================================== */
/* ==================================================== */

GLOBALFUN void PlayBeat(FILE *fp, INT32 pattern, INT32 beat, INT32 bpms )
{
    INT32 length = SAMPLE_FREQUENCY * 60 / bpms+1;

    BeatAssignDrumsToVoices(pattern,beat);

    while( length>0 )
    {
        if( length >= (16*1024) )
        {
            VoicesExtractTo32Bit(Global.Buffer32_l,Global.Buffer32_r,16*1024);
            Convert32To16Bit(Global.Buffer32_l,Global.Buffer32_r,Global.Buffer16,16*1024);
            if( fp )
                fwrite(Global.Buffer16,2*16*1024,sizeof(INT16),fp);
            else
                SoundcardWrite((char *)Global.Buffer16,2*2*16*1024);
            length -= 8*1024;
        }
        else
        {
            VoicesExtractTo32Bit(Global.Buffer32_l,Global.Buffer32_r,length);
            Convert32To16Bit(Global.Buffer32_l,Global.Buffer32_r,Global.Buffer16,length);
            if( fp )
                fwrite(Global.Buffer16,2*length,sizeof(INT16),fp);                
            else
                SoundcardWrite((char *)Global.Buffer16,2*2*length);
            length = 0;
        }
    }
}

/* ==================================================== */
/* ==================================================== */

GLOBALFUN void SongSaveAsSample(char *filename,INT32 start,INT32 stop)
{
    INT32 patt;
    INT32 beat;
    INT32 pos;
    INT32 voice;
    FILE *fp = fopen(filename, "w");

    printf("song: %s bpm: %d  start: %d stop %d beat\n",
           SONG_NAME,SONG_UPMS,start,stop);
    
    if( fp == 0 )
    {
        fprintf(stderr,"cannot open file %s\n",filename);
        return;
    }    
   
    FOREACH_VOICE(voice)
    {
        VoiceClear(voice);
    }
    
    for(pos=start;pos<=stop;pos++)
    {
        patt = SONG_MEASURE[pos];
        printf("processing pattern: %d[%s]\n",patt,PATT_NAME(patt));
        
        FOREACH_PATT_BEAT(patt,beat)
        {
            printf("patt: %d[%s] beat: %d\n",patt,PATT_NAME(patt),beat);
            PlayBeat(fp,patt,beat,SONG_UPMS*PATT_BEATS_PER_UNIT(patt));
        }
        
    }
    fclose(fp);
}

/* ==================================================== */
/* ==================================================== */

GLOBALFUN void PatternSaveAsSample(char *filename,INT32 patt,INT32 bmps)
{
    INT32 beat;
    INT32 voice;
    FILE *fp = fopen(filename, "w");

    if( fp == 0 )
    {
        fprintf(stderr,"cannot open file %s\n",filename);
        return;
    }    
    
    FOREACH_VOICE(voice)
    {
        VoiceClear(voice);
    }
    
    printf("processing pattern: %d[%s]\n",patt,PATT_NAME(patt));
        
    FOREACH_PATT_BEAT(patt,beat)
    {
        printf("patt: %d[%s] beat: %d\n",patt,PATT_NAME(patt),beat);
        PlayBeat(fp,patt,beat,SONG_UPMS*PATT_BEATS_PER_UNIT(patt));
    }
    
    fclose(fp);
}

/* ==================================================== */

/* ==================================================== */

GLOBALFUN void DrumLoad(INT32 drum, char *filename)
{
    FILE *fp = fopen(filename,"r");
    if( fp == 0 )
    {
        fprintf(stderr,"cannot load drum file %s\n",filename);
        strcpy(DRUM_PATH(drum),"NONE");
        DRUM_LENGTH(drum) = 0;
        return;
    }

    fprintf(stderr,"loading drum file %s\n",filename);
            
    strcpy(DRUM_PATH(drum),filename);
    fseek(fp,0,SEEK_END);
    DRUM_LENGTH(drum) = ftell(fp) / 2;
    fseek(fp,0,SEEK_SET);
    DRUM_SAMPLE(drum) = calloc(DRUM_LENGTH(drum),sizeof(INT16));
    fread(DRUM_SAMPLE(drum),sizeof(INT16),DRUM_LENGTH(drum),fp);
    fclose(fp);    
}


/* ==================================================== */

/* ==================================================== */ 

GLOBALFUN void SongSave(char *filename)
{
    int i,j,k;
    FILE *fp;
    
    fp = fopen(filename, "w" );
    
    if( fp == 0 )
    {
        fprintf(stderr,"cannot save file %s\n",filename);
        return;
    }

    fprintf(fp,
            "# \n"
            "# Songfile created by " TROMMLER_INFO "\n"
            "# " TROMMLER_COPYRIGHT "\n"
            "#\n");

    fprintf(fp,"v 3\n");   /* <<<<<<<< SONG FILE FORMAT VERSION >>>>>>>>> */
    
    fprintf(fp,
            "# \n"
            "# Drums\n"
            "#\n");

    for(i=0;i<MAX_DRUM_NUM;i++)
    {
        CHAR path[MAX_PATH_LEN];

        if( 0 == strncmp(DRUM_PATH(i),Global.DrumDir,strlen(Global.DrumDir)) )
        {
            strcpy(path,"@");
            strcat( path, DRUM_PATH(i) + strlen(Global.DrumDir) );
        }
        else
        {
            strcpy(path,DRUM_PATH(i));
        }
        
        fprintf(fp,"d %3d \"%s\" \"%s\" %3d %1.4f %1d %3d\n",
                i,DRUM_NAME(i),path,DRUM_VOLUME(i),DRUM_PAN(i),DRUM_MUTE(i),DRUM_KEY(i));
    }
    
    fprintf(fp,
            "# \n"
            "# Patterns\n"
            "#\n");

    for(i=0;i<MAX_PATT_NUM;i++)
    {
        if( PATT_UNITS(i) <= 0 ) continue;
        
        fprintf(fp,"p %3d \"%s\" %3d  %3d\n",i,PATT_NAME(i),PATT_UNITS(i),PATT_BEATS_PER_UNIT(i));
        for(j=0;j<MAX_DRUM_NUM;j++)
        {
            char buffer[128];
            
            for(k=0;k<PATT_BEATS(i);k++)
                if( PATT_DRUM_BEAT(i,j,k) )
                    break;

            if( k == PATT_BEATS(i) ) continue;

            for(k=0;k<PATT_BEATS(i);k++)
                if( PATT_DRUM_BEAT(i,j,k) )
                    buffer[k] = '*';
                else
                    buffer[k] = '.';
            buffer[k] = 0;

            fprintf(fp,"b %3d %3d %s\n",i,j,buffer);
        }
    }

    fprintf(fp,
            "# \n"
            "# Song\n"
            "#\n");

    fprintf(fp,"s \"%s\" %4d %4d\n",SONG_NAME,SONG_UPMS,SONG_LENGTH);

    for(i=0;i<SONG_LENGTH;i++)
        fprintf(fp,"%d\n",SONG_MEASURE[i]);
               
    fprintf(fp,
            "#\n"
            "# eof\n"
            "#\n");

    fclose(fp);
}


/* ==================================================== */

/* ==================================================== */ 

GLOBALFUN void SongLoad(char *filename)
{
    FILE *fp;
    char buffer[4*1024];
    CHAR beats[MAX_BEAT_NUM];
    CHAR name[MAX_NAME_LEN];
    CHAR path[MAX_PATH_LEN];
    INT32 patt;
    INT32 drum;
    INT32 beat;
    INT32 units;
    INT32 bpu;
    INT32 vol;
    INT32 key;
    float pan;
    INT32 mute;
    INT32 upm;
    INT32 len;
    CHAR dummy;

    /* old song files did not have a version number so we default to 1 */
    INT32 version = 1;   

    /* ==================================================== */

    ClearAll();

   /* ==================================================== */

   printf("++ Loading Song %s\n",filename);
    
   fp = fopen(filename, "r" );
   
   if( fp == 0 )
   {
       fprintf(stderr,"cannot load file %s\n",filename);
       return;
   }
   
   while(!feof(fp))
   {
       fgets(buffer,sizeof(buffer),fp);

       switch( buffer[0] )
       {
       case '\n':
       case '\t':
       case ' ':
       case '#':
           break;

       case 'v':
           sscanf(buffer,"%c %d",
                  &dummy,&version);
           break;
           
       case 'p':
           sscanf(buffer,"%c %d \"%[^\"]\" %d %d",
                  &dummy,&patt,name,&units, &bpu);
#if 0
           fprintf(stderr,"patt: %d[%s]  units: %d\n",patt,name,units);
#endif
           strcpy(PATT_NAME(patt),name);
           PATT_UNITS(patt) = units;
           PATT_BEATS_PER_UNIT(patt) = bpu;
           break;

       case 'b':
           sscanf(buffer,"%c %d %d %s",
                  &dummy,&patt,&drum,beats);
#if 0
           fprintf(stderr,"patt: %d  drum: %d  beats: %s\n",patt,drum,beats);
#endif
           FOREACH_PATT_BEAT(patt,beat)
           {
               if( beats[beat] == '*' )
                   PATT_DRUM_BEAT(patt,drum,beat) = 1;
               else
                   PATT_DRUM_BEAT(patt,drum,beat) = 0;
           }
           break;

       case 'd':
           /* there appears to be a problem with empty path */

           path[0] = '\0';
           
           if( version < 2 )
           {
               sscanf(buffer,"%c %d \"%[^\"]\" \"%[^\"]\" %d %d",
                      &dummy,&drum,name,path,&vol,&key);
               pan = 0.0;
               mute = 0;
           }
           else if (version < 3)
           {
                sscanf(buffer,"%c %d \"%[^\"]\" \"%[^\"]\" %d %f %d",
                      &dummy,&drum,name,path,&vol,&pan,&key);
                mute = 0;
           }
           else
           {
                sscanf(buffer,"%c %d \"%[^\"]\" \"%[^\"]\" %d %f %d %d",
                      &dummy,&drum,name,path,&vol,&pan,&mute,&key);
           }

           strcpy(DRUM_NAME(drum),name);
           if( path[0] == '@' )
           {
               strcpy(DRUM_PATH(drum),Global.DrumDir);
               strcat(DRUM_PATH(drum),path+1);
           }
           else
           {
               strcpy(DRUM_PATH(drum),path);
           }

           if( drum >= MAX_DRUM_NUM )
           {
               fprintf(stderr,"drum index to large:  %d [%s]  path: %s size: %d  pan %f\n",
                       drum,DRUM_NAME(drum),DRUM_PATH(drum),DRUM_LENGTH(drum),DRUM_PAN(drum));
           }
           else
           {
               DRUM_VOLUME(drum) = vol;
               DRUM_PAN(drum)   = pan;
               DRUM_MUTE(drum)  = mute;
               DRUM_KEY(drum)   = key;
               DrumLoad(drum,DRUM_PATH(drum));

#if 0
               fprintf(stderr,"drum: %d [%s]  path: %s size: %d  pan %f\n",
                   drum,DRUM_NAME(drum),DRUM_PATH(drum),DRUM_LENGTH(drum),DRUM_PAN(drum));
#endif
           

           }
           
           break;

       case 's':
           sscanf(buffer,"%c \"%[^\"]\" %d %d",
                  &dummy,name,&upm,&len);

           SONG_UPMS = upm;
           strcpy(SONG_NAME,name);
           if( len >= MAX_MEASURE_NUM )
           {
               fprintf(stderr,"song too long: %s bpm %d len: %d\n",name,upm,len);
               len = 0;
           }

           SONG_LENGTH = len;

#if 0
           fprintf(stderr,"song: %s bpm %d len: %d\n",SONG_NAME,SONG_BPMS,SONG_LENGTH);
#endif      
           for(patt=0; patt < len; patt++ )
           {
               INT32 pattern;
               
               fgets(buffer,sizeof(buffer),fp);
               sscanf(buffer,"%d",&pattern);
               if( pattern >= MAX_PATT_NUM )
               {
                   fprintf(stderr,"pattern index %d to large at measure:  %d\n",pattern,patt);
                   pattern = 0;
               }

               SONG_MEASURE[patt] = pattern;
           }
           break;
       default:
           fprintf(stderr,"bad line in songfile: %s\n",filename);
           fprintf(stderr,buffer);
           break;
       }
   }

   strncpy(SONG_PATH, filename, MAX_PATH_LEN);
   
   printf("-- Done version %d\n",version);
}

/* ==================================================== */
/* ==================================================== */





