/*
    gui.c
    
    Trommler - X11 Drum Machine
    Copyright (C) 1998,2002 Robert Muth <trommler@muth.org>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; version 2 of June 1991.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program in form of the file COPYING; 
    if not, write to the 

    Free Software Foundation, Inc. <http://www.fsf.org>
    59 Temple Place, Suite 330, 
    Boston, MA 02111-1307  USA

    $Id: gui.c,v 1.6 2002/11/30 16:59:28 muth Exp $
*/

#include <gtk/gtk.h>
#include <gdk/gdk.h>
#include <gdk/gdkkeysyms.h>
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>

#include "std.h"
#include "drummer.h"
#include "version.h"
#include "export.h"

#undef DEBUG


/* ==================================================== */
/* ==================================================== */

LOCALTYPE typedef enum
{
    PLAY_INVALID,
    PLAY_PATTERN,
    PLAY_RANGE,
    PLAY_SONG
}TYPE_PLAY;


/* ==================================================== */
/* ==================================================== */

LOCALVAR struct
{
    int StartMeasure;
    int StopMeasure;
    int Pattern;
  
    TYPE_PLAY PlayMode;
    int PlayPattern;    
    int PlayBeat;
    int PlayMeasure;

    int PlayLoop;
    int PlayActive;    
}GuiCurrent;

/* ==================================================== */
/* ==================================================== */

LOCALVAR struct
{
    GtkStyle *ButtonEven;
    GtkStyle *ButtonOdd;
    
    GtkTooltips *Tooltips;

    GtkWidget *Toplevel;
    
    GtkWidget *PatternNumber;
    GtkWidget *PatternName;
    GtkWidget *CopyPatternName;
    GtkWidget *PatternSource;
    GtkWidget *PatternUnits;
    GtkWidget *PatternBeatsPerUnit;

    GtkWidget *DrumContainer;
    GtkWidget *DrumName[MAX_DRUM_NUM];
    GtkWidget *DrumVolume[MAX_DRUM_NUM];
    GtkWidget *DrumPan[MAX_DRUM_NUM];
    GtkWidget *DrumMute[MAX_DRUM_NUM];
    GtkWidget *DrumBeat[MAX_DRUM_NUM][MAX_BEAT_NUM];                

    GtkWidget *SongFile;
    GtkWidget *SongName;
    GtkWidget *SongMeasures;
    GtkWidget *SongBpms;

    
    GtkWidget *SongMeasureBox[MAX_MEASURE_NUM];
    GtkWidget *SongMeasurePattern[MAX_MEASURE_NUM];
    GtkWidget *SongRangeStart;
    GtkWidget *SongRangeStop;

    GtkWidget *PlayDisplay;
    GtkWidget *PlayActiveButton;
    GtkWidget *PlayLoopButton;
}GuiWidgets;


/* ==================================================== */
/* ==================================================== */
/* ==================================================== */
/* ==================================================== */

LOCALFUN void SetPlay()
{
    char buffer[64];
    sprintf(buffer,"%03d:%03d:%03d",
            GuiCurrent.PlayMeasure,
            GuiCurrent.PlayPattern,
            GuiCurrent.PlayBeat);
    
    gtk_label_set(GTK_LABEL(GuiWidgets.PlayDisplay),buffer);
}
/* ==================================================== */
/* ==================================================== */

LOCALFUN void SetDrumMute(int drum)
{
    gtk_toggle_button_set_state( GTK_TOGGLE_BUTTON(GuiWidgets.DrumMute[drum]),(gfloat)DRUM_MUTE(drum));
}

/* ==================================================== */
/* ==================================================== */

LOCALFUN void SetDrumPan(int drum)
{
    gtk_spin_button_set_value( GTK_SPIN_BUTTON(GuiWidgets.DrumPan[drum]),(gfloat)DRUM_PAN(drum));
}

/* ==================================================== */
/* ==================================================== */

LOCALFUN void SetDrumVolume(int drum)
{
    gtk_spin_button_set_value( GTK_SPIN_BUTTON(GuiWidgets.DrumVolume[drum]),(gfloat)DRUM_VOLUME(drum));
}

/* ==================================================== */
/* ==================================================== */

LOCALFUN void SetDrumName(int drum)
{
    gtk_entry_set_text(GTK_ENTRY(GuiWidgets.DrumName[drum]),DRUM_NAME(drum));
}

/* ==================================================== */
/* ==================================================== */

LOCALFUN void SetDrums()
{
    int i;
    for(i=0;i<MAX_DRUM_NUM;i++)
    {
        SetDrumName(i);
        SetDrumVolume(i);
        SetDrumPan(i);
        SetDrumMute(i);
    }
}

/* ==================================================== */
/* ==================================================== */

LOCALFUN void SetPatternLength(void)
{
    
    int i,j;
    int pattern =  GuiCurrent.Pattern;

    gtk_spin_button_set_value( GTK_SPIN_BUTTON(GuiWidgets.PatternBeatsPerUnit),(gfloat) PATT_BEATS_PER_UNIT(pattern));
    gtk_spin_button_set_value( GTK_SPIN_BUTTON(GuiWidgets.PatternUnits),(gfloat) PATT_UNITS(pattern));
    
    for(j=0;j<MAX_BEAT_NUM;j++)
    for(i=0;i<MAX_DRUM_NUM;i++)
    {
        GtkWidget *widget = GuiWidgets.DrumBeat[i][j];

        if( (j / PATT_BEATS_PER_UNIT(pattern)) & 1 )
            gtk_widget_set_style(widget,GuiWidgets.ButtonOdd);
        else
            gtk_widget_set_style(widget,GuiWidgets.ButtonEven);

#if 1                
        if( PATT_DRUM_BEAT(pattern,i,j) )
        {
            gtk_toggle_button_set_state(GTK_TOGGLE_BUTTON(widget),1);
            if( GTK_TOGGLE_BUTTON(widget)->active==0 )
                printf("BAD BUTTON BEHAVIOUR\n");
        }
        else
        {
            gtk_toggle_button_set_state(GTK_TOGGLE_BUTTON(widget),0);
            if( GTK_TOGGLE_BUTTON(widget)->active!=0 )
                printf("BAD BUTTON BEHAVIOUR\n");
        }
#endif
        
#if 0
        if( j < PATT_BEATS(pattern) )
            gtk_widget_set_sensitive(widget,1);
        else
            gtk_widget_set_sensitive(widget,0);
#else
        if( j < PATT_BEATS(pattern) )
            gtk_widget_show(widget);
        else
            gtk_widget_hide(widget);
#endif
    }


}

/* ==================================================== */
/* ==================================================== */

LOCALFUN void SetPatternNumber()
{
    gtk_spin_button_set_value( GTK_SPIN_BUTTON(GuiWidgets.PatternNumber),(gfloat)GuiCurrent.Pattern);
}

/* ==================================================== */
/* ==================================================== */

LOCALFUN void SetPatternName()
{
    gtk_entry_set_text(GTK_ENTRY(GuiWidgets.PatternName),PATT_NAME(GuiCurrent.Pattern));
}

/* ==================================================== */
/* ==================================================== */

LOCALFUN void SetPattern()
{
    SetPatternName();
    SetPatternNumber();
    SetPatternLength();    
}

/* ==================================================== */
/* ==================================================== */

LOCALFUN void SetSongLength(INT32 old_song_length)
{
    INT32 i;
    gchar buffer[16];

    gtk_spin_button_set_value( GTK_SPIN_BUTTON(GuiWidgets.SongMeasures),(gfloat)SONG_LENGTH);

    for(i=0;i<MAX_MEASURE_NUM;i++)
    {
        if ( i >= old_song_length && i < SONG_LENGTH )
        {
            sprintf(buffer, "%3d", SONG_MEASURE[i]);
            gtk_entry_set_text(GTK_ENTRY(GuiWidgets.SongMeasurePattern[i]),buffer);
        }

        if( i < SONG_LENGTH )
            gtk_widget_show(GuiWidgets.SongMeasureBox[i]);
        else
            gtk_widget_hide(GuiWidgets.SongMeasureBox[i]);
    }
}

/* ==================================================== */
/* ==================================================== */

LOCALFUN void SetSongName()
{
    gtk_entry_set_text(GTK_ENTRY(GuiWidgets.SongName),SONG_NAME);
}

/* ==================================================== */
/* ==================================================== */

LOCALFUN void SetSongPath()
{
    gtk_label_set(GTK_LABEL(GuiWidgets.SongFile),SONG_PATH);
}

/* ==================================================== */
/* ==================================================== */

LOCALFUN void SetSongBpms()
{
    gtk_spin_button_set_value( GTK_SPIN_BUTTON(GuiWidgets.SongBpms),(gfloat)SONG_UPMS);
}

/* ==================================================== */
/* ==================================================== */

LOCALFUN void SetSongMeasure(pos)
{
    gchar buffer[16];

    sprintf(buffer,"%3d",SONG_MEASURE[pos]);
    gtk_entry_set_text(GTK_ENTRY(GuiWidgets.SongMeasurePattern[pos]),buffer);
}
/* ==================================================== */
/* ==================================================== */

LOCALFUN void SetSong()
{
    int i;
    
    for(i=0;i<MAX_MEASURE_NUM;i++)
        SetSongMeasure(i);

    SetSongLength(0);
    SetSongBpms();
    SetSongName();
    SetSongPath();
    GuiCurrent.Pattern = 0;
    SetPattern();
    SetDrums();
    
    /* SetPlaycurr(); */
}

/* ==================================================== */
/* ==================================================== */
LOCALFUN int ActionPlayReset()
{
    GuiCurrent.PlayMeasure = 0; 
    GuiCurrent.PlayPattern = 0;
    GuiCurrent.PlayBeat = 0;
    SetPlay();
    return TRUE;
}
/* ==================================================== */
/* ==================================================== */

LOCALFUN int ActionPlayLoop()
{
    GuiCurrent.PlayLoop = GTK_TOGGLE_BUTTON (GuiWidgets.PlayLoopButton)->active;
    return TRUE;
}

/* ==================================================== */
/* ==================================================== */
LOCALFUN int ActionPlayActive()
{
    GuiCurrent.PlayActive = GTK_TOGGLE_BUTTON (GuiWidgets.PlayActiveButton)->active;
    return TRUE;
}
/* ==================================================== */
/* ==================================================== */
LOCALFUN int ActionPlayPattern()
{
    GuiCurrent.PlayMode = PLAY_PATTERN;
    return TRUE;
}
/* ==================================================== */
/* ==================================================== */
LOCALFUN int ActionPlayRange()
{
    GuiCurrent.PlayMode = PLAY_RANGE;
    return TRUE;
}
/* ==================================================== */
/* ==================================================== */
LOCALFUN int ActionPlaySong()
{
    GuiCurrent.PlayMode = PLAY_SONG;
    return TRUE;
}
/* ==================================================== */
/* ==================================================== */
LOCALFUN int ActionCopyPattern()
{
    GtkAdjustment *adjustment = gtk_spin_button_get_adjustment( GTK_SPIN_BUTTON(GuiWidgets.PatternSource));

    PattCopy(GuiCurrent.Pattern,(int)adjustment->value);
    SetPattern();
    return TRUE;
}
/* ==================================================== */
/* ==================================================== */
LOCALFUN int ActionBeatsRotateRight(GtkWidget *w)
{
    int drum;
    int beat;
    int pattern = GuiCurrent.Pattern;
    int tmp;
    
    for(drum=0;drum<MAX_DRUM_NUM;drum++)
    {
        tmp = PATT_DRUM_BEAT(pattern,drum,PATT_BEATS(pattern)-1);
        for(beat = PATT_BEATS(pattern)-1;beat > 0;beat--)
            PATT_DRUM_BEAT(pattern,drum,beat)=PATT_DRUM_BEAT(pattern,drum,beat-1);
        PATT_DRUM_BEAT(pattern,drum,0) = tmp;
    }
    
    SetPattern();
    return TRUE;
}
/* ==================================================== */
/* ==================================================== */
LOCALFUN int ActionBeatsRotateLeft(GtkWidget *w)
{
    int drum;
    int beat;
    int pattern = GuiCurrent.Pattern;
    int tmp;
    
    for(drum=0;drum<MAX_DRUM_NUM;drum++)
    {
        tmp = PATT_DRUM_BEAT(pattern,drum,0);
        for(beat = 0;beat < PATT_BEATS(pattern)-1;beat++)
            PATT_DRUM_BEAT(pattern,drum,beat)=PATT_DRUM_BEAT(pattern,drum,beat+1);
        PATT_DRUM_BEAT(pattern,drum,PATT_BEATS(pattern)-1) = tmp;
    }
    
    SetPattern();
    return TRUE;
}
/* ==================================================== */
/* ==================================================== */

LOCALFUN gint ActionToggleDrumBeat(GtkToggleButton *button)
{
    int drum = (int) gtk_object_get_data(GTK_OBJECT(button),"trommler_drum");
    int beat = (int) gtk_object_get_data(GTK_OBJECT(button),"trommler_beat");
    
    if( button->active )
        PATT_DRUM_BEAT(GuiCurrent.Pattern,drum,beat) = 1;
    else
        PATT_DRUM_BEAT(GuiCurrent.Pattern,drum,beat) = 0;
    return TRUE;
}

/* ==================================================== */
/* ==================================================== */

LOCALFUN gint ActionToggleDrumMute(GtkToggleButton *button)
{
    int drum = (int) gtk_object_get_data(GTK_OBJECT(button),"trommler_drum");
 
    if( button->active )
    {
        DRUM_MUTE(drum)     = 1;        /* mute on */
    }
    else
    {
        DRUM_MUTE(drum)     = 0;        /* mute off */
    }

    return TRUE;
}


/* ==================================================== */
/* ==================================================== */
LOCALFUN gint ActionChangeDrumVolume(GtkAdjustment *adjustment)
{
    int drum =  (int) gtk_object_get_data(GTK_OBJECT(adjustment),"trommler_drum");
    DRUM_VOLUME(drum) = adjustment->value;
    return TRUE;    
}

/* ==================================================== */
/* ==================================================== */
LOCALFUN gint ActionChangeDrumPan(GtkAdjustment *adjustment)
{
    int drum =  (int) gtk_object_get_data(GTK_OBJECT(adjustment),"trommler_drum");
    DRUM_PAN(drum) = ((float)(int)(adjustment->value*1000))/1000;
    gtk_adjustment_set_value(adjustment, DRUM_PAN(drum));
    return TRUE;    
}

/* ==================================================== */
/* ==================================================== */
LOCALFUN gint ActionChangePatternNumber(GtkAdjustment *adjustment)
{
    GuiCurrent.Pattern = adjustment->value;
    SetPatternName();
    SetPatternLength();
    return TRUE;    
}


/* ==================================================== */
/* ==================================================== */
LOCALFUN gint ActionChangeCopyPatternNumber(GtkAdjustment *adjustment)
{
    gtk_entry_set_text(GTK_ENTRY(GuiWidgets.CopyPatternName),PATT_NAME((INT32)adjustment->value));
    return TRUE;    
}

/* ==================================================== */
/* ==================================================== */

LOCALFUN gint ActionChangePatternNumber2(GtkObject *object)
{
    int measure = (int) gtk_object_get_data(object,"trommler_measure");

    GuiCurrent.Pattern = SONG_MEASURE[measure];
    SetPattern();
    return TRUE;    
}

/* ==================================================== */
/* ==================================================== */

LOCALFUN gint ActionChangePatternUnits(GtkAdjustment *adjustment)
{
    PATT_UNITS(GuiCurrent.Pattern) = adjustment->value;
    SetPatternLength();
#if 1
    usleep(400000);
#endif
    return TRUE;    
}

/* ==================================================== */
/* ==================================================== */

LOCALFUN gint ActionChangePatternBeatsPerUnit(GtkAdjustment *adjustment)
{
    PATT_BEATS_PER_UNIT(GuiCurrent.Pattern) = adjustment->value;
    SetPatternLength();
#if 1
    usleep(400000);
#endif
    return TRUE;    
}

/* ==================================================== */
/* ==================================================== */

LOCALFUN gint ActionChangeSongLength(GtkAdjustment *adjustment)
{
    INT32 old_song_length;

    old_song_length = SONG_LENGTH;
    SONG_LENGTH = adjustment->value;
    SetSongLength(old_song_length);
    return TRUE;    
}

/* ==================================================== */
/* ==================================================== */

LOCALFUN gint ActionChangeSongUPMS(GtkAdjustment *adjustment)
{
    SONG_UPMS = adjustment->value;
    return TRUE;    
}

/* ==================================================== */
/* ==================================================== */

LOCALFUN gint ActionRestoreMeasurePattern(GtkWidget *widget )
{
    int measure = (int) gtk_object_get_data(GTK_OBJECT(widget),"trommler_measure");
    SetSongMeasure(measure);
    return TRUE;    
}

/* ==================================================== */
/* ==================================================== */

LOCALFUN gint ActionChangeMeasurePattern(GtkObject *object )
{
    int measure = (int) gtk_object_get_data(object,"trommler_measure");
    int pattern = atoi( gtk_entry_get_text(GTK_ENTRY(GuiWidgets.SongMeasurePattern[measure])));
    SONG_MEASURE[measure] = pattern;
    return TRUE;    
}
/* ==================================================== */
/* ==================================================== */

LOCALFUN gint ActionRestoreDrumName(GtkObject *object )
{
    int drum = (int) gtk_object_get_data(object,"trommler_drum");
    SetDrumName(drum);
    return TRUE;    
}

/* ==================================================== */
/* ==================================================== */

LOCALFUN gint ActionChangeDrumName(GtkWidget *widget )
{
    int drum = (int) gtk_object_get_data(GTK_OBJECT(widget),"trommler_drum");
    gchar *name = gtk_entry_get_text(GTK_ENTRY(GuiWidgets.DrumName[drum]));
    strcpy( DRUM_NAME(drum),name);
    return TRUE;    
}
/* ==================================================== */
/* ==================================================== */

LOCALFUN gint ActionRestoreSongName()
{
    SetSongName();
    return TRUE;    
}
/* ==================================================== */
/* ==================================================== */

LOCALFUN gint ActionChangeSongName()
{
    gchar *songname = gtk_entry_get_text( GTK_ENTRY(GuiWidgets.SongName) );
    strcpy(SONG_NAME,songname);
    return TRUE;    
}
/* ==================================================== */
/* ==================================================== */

LOCALFUN gint ActionRestorePatternName()
{
    SetPatternName();
    return TRUE;    
}

/* ==================================================== */
/* ==================================================== */

LOCALFUN gint ActionChangePatternName()
{
    gchar *name = gtk_entry_get_text( GTK_ENTRY(GuiWidgets.PatternName) );
    strcpy(PATT_NAME(GuiCurrent.Pattern),name);
    return TRUE;    
}

/* ==================================================== */
/* ==================================================== */
LOCALFUN gint ActionChangeSongRangeStart(GtkWidget *widget)
{
    GuiCurrent.StartMeasure = GTK_ADJUSTMENT(widget)->value;
    return TRUE;    
}
/* ==================================================== */
/* ==================================================== */
LOCALFUN gint ActionSetStartMeasure(GtkWidget *widget)
{
    int measure = (int) gtk_object_get_data(GTK_OBJECT(widget),"trommler_measure");
    GuiCurrent.StartMeasure = measure;
    gtk_spin_button_set_value( GTK_SPIN_BUTTON(GuiWidgets.SongRangeStart),measure);
    
    return TRUE;    
}
/* ==================================================== */
/* ==================================================== */
LOCALFUN gint ActionChangeSongRangeStop(GtkWidget *widget)
{
    GuiCurrent.StopMeasure = GTK_ADJUSTMENT(widget)->value;
    return TRUE;    
}
/* ==================================================== */
/* ==================================================== */
LOCALFUN gint ActionSetStopMeasure(GtkWidget *widget)
{
    int measure = (int) gtk_object_get_data(GTK_OBJECT(widget),"trommler_measure");
    GuiCurrent.StartMeasure = measure;
    gtk_spin_button_set_value( GTK_SPIN_BUTTON(GuiWidgets.SongRangeStop),measure);

    return TRUE;    
}
/* ==================================================== */
/* ==================================================== */
LOCALFUN gint ActionDeleteMeasure(GtkWidget *widget)
{
    int measure = (int) gtk_object_get_data(GTK_OBJECT(widget),"trommler_measure");

    for( ;measure < MAX_MEASURE_NUM-1;measure++)
    {
        SONG_MEASURE[measure] = SONG_MEASURE[measure+1];
        SetSongMeasure(measure);        
    }

    return TRUE;    
}

/* ==================================================== */
/* ==================================================== */
LOCALFUN gint ActionDuplicateMeasure(GtkWidget *widget)
{
    int measure = (int) gtk_object_get_data(GTK_OBJECT(widget),"trommler_measure");
    int i;
    
#ifdef DEBUG
    printf("Duplicate Measure  %d\n",measure);
#endif
    
    for( i = MAX_MEASURE_NUM-1;i>measure;i--)
    {
        SONG_MEASURE[i] = SONG_MEASURE[i-1];
        SetSongMeasure(i);        
    }

    return TRUE;    
}

/* ==================================================== */
/* ==================================================== */
LOCALFUN gint ActionClearDrumBeats(GtkWidget *widget)
{
    int beat;
    int pattern = GuiCurrent.Pattern;
    int drum = (int) gtk_object_get_data(GTK_OBJECT(widget),"trommler_drum");

    for(beat=0;beat<MAX_BEAT_NUM;beat++)
    {
        GtkWidget *widget = GuiWidgets.DrumBeat[drum][beat];
        PATT_DRUM_BEAT(pattern,drum,beat)=0;
        gtk_toggle_button_set_state(GTK_TOGGLE_BUTTON(widget),0);
    }
                                        
    return TRUE;    
}
/* ==================================================== */
/* ==================================================== */

LOCALFUN int ActionDrumSwapFirst(GtkWidget *widget)
{
    int pattern;
    int beat;
    int drum = (int) gtk_object_get_data(GTK_OBJECT(widget),"trommler_drum");
    int k;
    
    TYPE_DRUM temp;

    temp = DRUM(drum);
    DRUM(drum) = DRUM(0);
    DRUM(0) = temp;

    for(pattern=0; pattern<MAX_PATT_NUM;pattern++)
        for(beat=0; beat<MAX_BEAT_NUM;beat++)
        {
            k = PATT_DRUM_BEAT(pattern,drum,beat);
            PATT_DRUM_BEAT(pattern,drum,beat) = PATT_DRUM_BEAT(pattern,0,beat);
            PATT_DRUM_BEAT(pattern,0,beat) = k;
        }

    SetPattern();
    SetDrums();

    return TRUE;    
}

/* ==================================================== */
/* ==================================================== */

LOCALFUN gint ActionSaveSamplePatternOk(GtkFileSelection *filesel)
{
    int pattern = (int) gtk_object_get_data(GTK_OBJECT(filesel),"trommler_pattern");
    gchar *filename = gtk_file_selection_get_filename(filesel);
    gchar *dirname = g_dirname(filename);
    strcpy(Global.SampleDir, dirname);
    g_free(dirname);

    PatternSaveAsSample(filename,pattern,SONG_UPMS);
    gtk_widget_destroy(GTK_WIDGET(filesel));
    return TRUE;    
}


LOCALFUN gint ActionSaveSamplePattern()
{
    char buffer[128];
    GtkWidget *filesel;

    sprintf(buffer,"Save Pattern %d Sample As",GuiCurrent.Pattern);
    filesel = gtk_file_selection_new(buffer);

    gtk_object_set_data(GTK_OBJECT(filesel),"trommler_pattern",
                        (gpointer) GuiCurrent.Pattern);

    gtk_signal_connect_object(GTK_OBJECT(GTK_FILE_SELECTION(filesel)->cancel_button),
                              "clicked",
                              GTK_SIGNAL_FUNC(gtk_widget_destroy),
                              GTK_OBJECT(filesel));
    gtk_signal_connect_object(GTK_OBJECT(GTK_FILE_SELECTION(filesel)->ok_button),
                              "clicked",
                              GTK_SIGNAL_FUNC(ActionSaveSamplePatternOk),
                              GTK_OBJECT(filesel));

    gtk_file_selection_complete(GTK_FILE_SELECTION(filesel), Global.SampleDir);
    gtk_file_selection_set_filename(GTK_FILE_SELECTION(filesel), "*.smp");

    gtk_widget_show(GTK_WIDGET(filesel));                                        
    return TRUE;    
}
/* ==================================================== */
/* ==================================================== */

LOCALFUN gint ActionSaveSampleRangeOk(GtkFileSelection *filesel)
{
    gchar *filename = gtk_file_selection_get_filename(filesel);
    gchar *dirname = g_dirname(filename);
    strcpy(Global.SampleDir, dirname);
    g_free(dirname);
    SongSaveAsSample(filename,GuiCurrent.StartMeasure,GuiCurrent.StopMeasure);
    gtk_widget_destroy(GTK_WIDGET(filesel));
    return TRUE;    
}


LOCALFUN gint ActionSaveSampleRange()
{
    GtkWidget *filesel;
    filesel = gtk_file_selection_new("Save Range Sample As");
    gtk_signal_connect_object(GTK_OBJECT(GTK_FILE_SELECTION(filesel)->cancel_button),
                              "clicked",
                              GTK_SIGNAL_FUNC(gtk_widget_destroy),
                              GTK_OBJECT(filesel));
    gtk_signal_connect_object(GTK_OBJECT(GTK_FILE_SELECTION(filesel)->ok_button),
                              "clicked",
                              GTK_SIGNAL_FUNC(ActionSaveSampleRangeOk),
                              GTK_OBJECT(filesel));
    gtk_file_selection_complete(GTK_FILE_SELECTION(filesel), Global.SampleDir);
    gtk_file_selection_complete(GTK_FILE_SELECTION(filesel),"*.smp");
    gtk_widget_show(GTK_WIDGET(filesel));    
    return TRUE;    
}
/* ==================================================== */
/* ==================================================== */

LOCALFUN gint ActionSaveSampleSongOk(GtkFileSelection *filesel)
{
    gchar *filename = gtk_file_selection_get_filename(filesel);
    gchar *dirname = g_dirname(filename);
    strcpy(Global.SampleDir, dirname);
    g_free(dirname);
    SongSaveAsSample(filename,0,SONG_LENGTH-1);
    gtk_widget_destroy(GTK_WIDGET(filesel));
    return TRUE;    
}


LOCALFUN gint ActionSaveSampleSong()
{
    GtkWidget *filesel;
    filesel = gtk_file_selection_new("Save Song Sample As");
    gtk_signal_connect_object(GTK_OBJECT(GTK_FILE_SELECTION(filesel)->cancel_button),
                              "clicked",
                              GTK_SIGNAL_FUNC(gtk_widget_destroy),
                              GTK_OBJECT(filesel));
    gtk_signal_connect_object(GTK_OBJECT(GTK_FILE_SELECTION(filesel)->ok_button),
                              "clicked",
                              GTK_SIGNAL_FUNC(ActionSaveSampleSongOk),
                              GTK_OBJECT(filesel));
    gtk_file_selection_complete(GTK_FILE_SELECTION(filesel), Global.SampleDir);
    gtk_file_selection_complete(GTK_FILE_SELECTION(filesel),"*.smp");
    gtk_widget_show(GTK_WIDGET(filesel));    
    return TRUE;    
}
/* ==================================================== */
/* ==================================================== */
LOCALVAR int GlobalSetName = 0; // fixme ugly global var 

LOCALFUN gint ActionLoadDrumOk(GtkFileSelection *filesel)
{
    int drum = (int) gtk_object_get_data(GTK_OBJECT(filesel),"trommler_drum");
    gchar *filename = gtk_file_selection_get_filename(filesel);

#ifdef DEBUG
    printf("load drum %d %s\n",drum,filename);
#endif

    DrumLoad(drum,filename);

#ifdef DEBUG
    printf("destroying fileselector\n");
#endif
    
    gtk_widget_destroy(GTK_WIDGET(filesel));

#ifdef DEBUG
    printf("destroying fileselector done\n");
#endif

    if( GlobalSetName )
    {
        gchar *name = strrchr(filename,'/');
        if( name )
        {
            name++;
        }
        else
        {
            name = filename;
        }
        
        gtk_entry_set_text(GTK_ENTRY(GuiWidgets.DrumName[drum]),name);
        strcpy( DRUM_NAME(drum),name);
    }
    
            
    return TRUE;    
}


LOCALFUN gint ActionLoadDrumGeneric(GtkWidget *widget, int set_name)
{
    char buffer[128];
    int drumindex; 
    GtkWidget *filesel;

    GlobalSetName = set_name;
    
    drumindex = (int) gtk_object_get_data(GTK_OBJECT(widget),"trommler_drum");
    sprintf(buffer,"Load Drum %d",drumindex);
    filesel = gtk_file_selection_new(buffer);

    gtk_object_set_data(GTK_OBJECT(filesel),"trommler_drum",(gpointer) drumindex);

    gtk_signal_connect_object(GTK_OBJECT(GTK_FILE_SELECTION(filesel)->cancel_button),
                              "clicked",
                              GTK_SIGNAL_FUNC(gtk_widget_destroy),
                              GTK_OBJECT(filesel));
    gtk_signal_connect_object(GTK_OBJECT(GTK_FILE_SELECTION(filesel)->ok_button),
                              "clicked",
                              GTK_SIGNAL_FUNC(ActionLoadDrumOk),
                              GTK_OBJECT(filesel));


    gtk_file_selection_complete(GTK_FILE_SELECTION(filesel),Global.DrumDir);
    gtk_file_selection_set_filename(GTK_FILE_SELECTION(filesel),"*.smp");
    
    gtk_widget_show(GTK_WIDGET(filesel));    
                                    
    return TRUE;    
}

LOCALFUN gint ActionLoadDrumSetName(GtkWidget *widget)
{
    return ActionLoadDrumGeneric(widget,1);
}

LOCALFUN gint ActionLoadDrum(GtkWidget *widget)
{
    return ActionLoadDrumGeneric(widget,0);
}

/* ==================================================== */
/* ==================================================== */

LOCALFUN gint ActionSaveSong()
{
    SongSave(SONG_PATH);
    
    return TRUE;
}

/* ==================================================== */
/* ==================================================== */

LOCALFUN gint ActionSaveSongOk(GtkFileSelection *filesel)
{
    gchar *filename = gtk_file_selection_get_filename(filesel);
    strncpy(SONG_PATH, filename, MAX_PATH_LEN);
    SetSongPath();

#ifdef DEBUG
    printf("save song as %s\n",filename);
#endif

    SongSave(filename);
    gtk_widget_destroy(GTK_WIDGET(filesel));
    return TRUE;    
}

/* ==================================================== */
/* ==================================================== */

LOCALFUN gint ActionSaveSongAs()
{
    GtkWidget *widget = gtk_file_selection_new("Save Song");
    gtk_signal_connect_object(GTK_OBJECT(GTK_FILE_SELECTION(widget)->cancel_button),
                              "clicked",
                              GTK_SIGNAL_FUNC(gtk_widget_destroy),
                              GTK_OBJECT(widget));
    gtk_signal_connect_object(GTK_OBJECT(GTK_FILE_SELECTION(widget)->ok_button),
                              "clicked",
                              GTK_SIGNAL_FUNC(ActionSaveSongOk),
                              GTK_OBJECT(widget));

    gtk_file_selection_set_filename(GTK_FILE_SELECTION(widget),"*.sng");

    gtk_widget_show(GTK_WIDGET(widget));    
                                    
    return TRUE;    
}

/* ==================================================== */
/* ==================================================== */

LOCALFUN gint ActionNewSong()
{
    gchar *filename = "no_name.sng";
    gtk_label_set(GTK_LABEL(GuiWidgets.SongFile),filename);
#ifdef DEBUG
    printf("new song %s\n",filename);
#endif
    SongClear();    
    SetSong();    
    return TRUE;    
}

/* ==================================================== */
/* ==================================================== */

LOCALFUN gint ActionLoadSongOk(GtkFileSelection *filesel)
{
    gchar *filename = gtk_file_selection_get_filename(filesel);
    gtk_label_set(GTK_LABEL(GuiWidgets.SongFile),filename);
#ifdef DEBUG
    printf("load song %s\n",filename);
#endif
    SongLoad(filename);
    SetSong();
    
    gtk_widget_destroy(GTK_WIDGET(filesel));
    return TRUE;    
}

/* ==================================================== */
/* ==================================================== */

LOCALFUN gint ActionLoadSong()
{
    GtkWidget *widget=gtk_file_selection_new("Load Song");
    gtk_signal_connect_object(GTK_OBJECT(GTK_FILE_SELECTION(widget)->cancel_button),
                              "clicked",
                              GTK_SIGNAL_FUNC(gtk_widget_destroy),
                              GTK_OBJECT(widget));
    gtk_signal_connect_object(GTK_OBJECT(GTK_FILE_SELECTION(widget)->ok_button),
                              "clicked",
                              GTK_SIGNAL_FUNC(ActionLoadSongOk),
                              GTK_OBJECT(widget));

    gtk_file_selection_set_filename(GTK_FILE_SELECTION(widget),"*.sng");

    gtk_widget_show(GTK_WIDGET(widget));    
                                    
    return TRUE;    
}


/* ==================================================== */
/* ==================================================== */

LOCALFUN gint ActionQuit(GtkWidget *widget,GdkEvent *event)
{
    gtk_main_quit();
    return 0;
}

/* ==================================================== */
/* ==================================================== */

LOCALFUN gint ActionAbout()
{
    GtkWidget *dialog;
    GtkWidget *widget;

    dialog = gtk_dialog_new();
    
    widget = gtk_label_new(TROMMLER_ABOUT);
    gtk_box_pack_start( GTK_BOX(GTK_DIALOG(dialog)->vbox), widget, FALSE, FALSE, 0 );
    gtk_widget_show(widget);

    widget = gtk_button_new_with_label(" Ok ");
    gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->action_area),
                       widget,TRUE,FALSE,0);
    gtk_signal_connect_object(GTK_OBJECT(widget), "clicked",
                GTK_SIGNAL_FUNC(gtk_widget_destroy), GTK_OBJECT(dialog));

    gtk_widget_show(widget);

    gtk_window_position(GTK_WINDOW(dialog),GTK_WIN_POS_MOUSE);
    gtk_widget_grab_focus(dialog);
    gtk_widget_show(dialog);
    
    return TRUE;    
}


/* ==================================================== */
/* ==================================================== */

LOCALFUN gint ActionMeasurePopup(GtkWidget *widget,GdkEvent *event)
{
    LOCALFUN GtkWidget *menu=NULL;
    
    GtkWidget *item;
    GdkEventButton *bevent = (GdkEventButton *) event;
    int measure = (int) gtk_object_get_data(GTK_OBJECT(widget),"trommler_measure");

    if( event->type != GDK_BUTTON_PRESS ) return FALSE;
    
#ifdef DEBUG
    printf("popup for measure %d\n",measure);
#endif
    
    if( menu == 0)
    {
        menu = gtk_menu_new();

        item = gtk_menu_item_new_with_label("Show Pattern");
        gtk_signal_connect_object(GTK_OBJECT(item),
                                  "activate",
                                  GTK_SIGNAL_FUNC(ActionChangePatternNumber2),
                                  GTK_OBJECT(menu));
        gtk_menu_append( GTK_MENU(menu), item);
        gtk_widget_show(item);

        item = gtk_menu_item_new_with_label("Duplicate Measure");
        gtk_menu_append( GTK_MENU(menu), item);
        gtk_signal_connect_object(GTK_OBJECT(item),
                                  "activate",
                                  GTK_SIGNAL_FUNC(ActionDuplicateMeasure),
                                  GTK_OBJECT(menu));
        gtk_widget_show(item);

        item = gtk_menu_item_new_with_label("Delete Measure");
        gtk_menu_append( GTK_MENU(menu), item);
        gtk_signal_connect_object(GTK_OBJECT(item),
                                  "activate",
                                  GTK_SIGNAL_FUNC(ActionDeleteMeasure),
                                  GTK_OBJECT(menu));
        gtk_widget_show(item);

        item = gtk_menu_item_new_with_label("Set Range Start");
        gtk_menu_append( GTK_MENU(menu), item);
        gtk_signal_connect_object(GTK_OBJECT(item),
                                  "activate",
                                  GTK_SIGNAL_FUNC(ActionSetStartMeasure),
                                  GTK_OBJECT(menu));
        gtk_widget_show(item);

        item = gtk_menu_item_new_with_label("Set Range End");
        gtk_menu_append( GTK_MENU(menu), item);
        gtk_signal_connect_object(GTK_OBJECT(item),
                                  "activate",
                                  GTK_SIGNAL_FUNC(ActionSetStopMeasure),
                                  GTK_OBJECT(menu));
        gtk_widget_show(item);
    }

    gtk_object_set_data(GTK_OBJECT(menu),"trommler_measure",(gpointer) measure);

    gtk_menu_popup(GTK_MENU(menu),NULL,NULL,NULL,NULL,bevent->button,bevent->time);
    return TRUE;
}

/* ==================================================== */
/* ==================================================== */

LOCALFUN gint ActionDrumPopup(GtkWidget *widget,GdkEvent *event)
{
    static GtkWidget *menu=NULL;
    static GtkWidget *label=NULL;
        
#if 0
    static GtkWidget *file=NULL;
#endif
    GtkWidget *item;
    
    GdkEventButton *bevent = (GdkEventButton *) event;
    int drum = (int) gtk_object_get_data(GTK_OBJECT(widget),"trommler_drum");

    if( event->type != GDK_BUTTON_PRESS ) return FALSE;

#ifdef DEBUG
    printf("popup for drum %d\n",drum);    
#endif
        

    if( menu == 0)
    {
        menu = gtk_menu_new();

        item = gtk_menu_item_new_with_label("*No Path Yet*");
        gtk_menu_append( GTK_MENU(menu), item);
        label = item;
        
        item = gtk_menu_item_new_with_label("Load");
        gtk_menu_append( GTK_MENU(menu), item);
        gtk_signal_connect_object(GTK_OBJECT(item),
                                  "activate",
                                  GTK_SIGNAL_FUNC(ActionLoadDrum),
                                  GTK_OBJECT(menu));
        gtk_widget_show(item);

        item = gtk_menu_item_new_with_label("Load And Set Name");
        gtk_menu_append( GTK_MENU(menu), item);
        gtk_signal_connect_object(GTK_OBJECT(item),
                                  "activate",
                                  GTK_SIGNAL_FUNC(ActionLoadDrumSetName),
                                  GTK_OBJECT(menu));
        gtk_widget_show(item);

        item = gtk_menu_item_new_with_label("Clear");
        gtk_menu_append( GTK_MENU(menu), item);
        gtk_signal_connect_object(GTK_OBJECT(item),
                                  "activate",
                                  GTK_SIGNAL_FUNC(ActionClearDrumBeats),
                                  GTK_OBJECT(menu));
        gtk_widget_show(item);

        item = gtk_menu_item_new_with_label("Swap First");
        gtk_menu_append( GTK_MENU(menu), item);
        gtk_signal_connect_object(GTK_OBJECT(item),
                                  "activate",
                                  GTK_SIGNAL_FUNC(ActionDrumSwapFirst),
                                  GTK_OBJECT(menu));
        gtk_widget_show(item);

#if 0
        file = gtk_menu_item_new_with_label(buffer);
        gtk_menu_append( GTK_MENU(menu), file);
        gtk_widget_show(file);
#endif
    }

    /* show the path for the drum sample */
       
    /* gtk_label_set(GTK_LABEL(label),DRUM_PATH(drum)); */
    
    gtk_object_set_data(GTK_OBJECT(menu),"trommler_drum",(gpointer) drum);
    
    gtk_menu_popup(GTK_MENU(menu),NULL,NULL,NULL,NULL,bevent->button,bevent->time);
    return TRUE;
}

/* ==================================================== */
/* ==================================================== */

LOCALFUN gint KeyboardCallback(GtkWidget* widget, GdkEventKey* event, gpointer data)
{
    gint key = event->keyval;
#if 0
    printf("%04x\n",key);
#endif
    if( key >= GDK_KP_0 && key <=  GDK_KP_9 )
    {
        key -= GDK_KP_0;
        
        if( GuiCurrent.PlayMode == PLAY_PATTERN )
        {
            PATT_DRUM_BEAT(GuiCurrent.PlayPattern,key,GuiCurrent.PlayBeat) = 1;
            gtk_toggle_button_set_state(GTK_TOGGLE_BUTTON(GuiWidgets.DrumBeat[key][GuiCurrent.PlayBeat]), 1);
        }
    }
    return TRUE;
}


/* ==================================================== */
/* ==================================================== */
/* ==================================================== */
/* ==================================================== */
LOCALFUN void CreateMeasures(GtkWidget *container)
{
    int i;
    GtkWidget *widget = gtk_hbox_new(FALSE,0);
    /* gtk_container_add(GTK_CONTAINER(container), widget); */
    gtk_scrolled_window_add_with_viewport(
        GTK_SCROLLED_WINDOW(container), widget);
    /* Beat Items */
    for (i=0;i<MAX_MEASURE_NUM;i++)  
    {
        GtkWidget *parent = widget;
        GtkWidget *widget = gtk_vbox_new(FALSE,0);
        GuiWidgets.SongMeasureBox[i] = widget;
        
        gtk_box_pack_start( GTK_BOX(parent), widget, FALSE,FALSE,0);        
        {
            GtkWidget *parent = widget;
            GtkWidget *widget;
            char buffer[16];
            sprintf(buffer,"%03d",i);

            /* popup menu button */
            
            widget = gtk_button_new_with_label(buffer);
            gtk_object_set_data(GTK_OBJECT(widget),"trommler_measure",(gpointer) i);
            gtk_box_pack_start(GTK_BOX(parent),widget,FALSE,FALSE,0);
            gtk_signal_connect_object( GTK_OBJECT(widget),"event",
                                       GTK_SIGNAL_FUNC(ActionMeasurePopup),
                                       GTK_OBJECT(widget));
            
            gtk_tooltips_set_tip(GuiWidgets.Tooltips,
                                 widget,
                                 "Measure Popup Menu",
                                 "");

            /* measure change entry */
            
            widget = gtk_entry_new();
            GuiWidgets.SongMeasurePattern[i] = widget;
            gtk_object_set_data(GTK_OBJECT(widget),"trommler_measure",(gpointer) i);
            gtk_box_pack_start(GTK_BOX(parent),widget,FALSE,FALSE,0);
            gtk_entry_set_max_length(GTK_ENTRY(widget),3);
            gtk_entry_set_text(GTK_ENTRY(widget),"???");
            gtk_entry_set_editable(GTK_ENTRY(widget),TRUE);
            gtk_widget_set_usize(widget,30,-1);
            gtk_signal_connect_object( GTK_OBJECT(widget),
                                       "activate",
                                       GTK_SIGNAL_FUNC(ActionChangeMeasurePattern),
                                       (gpointer) widget);
            gtk_signal_connect_object( GTK_OBJECT(widget),
                                       "focus_out_event",
                                       GTK_SIGNAL_FUNC(ActionChangeMeasurePattern),
                                       (gpointer) widget);
        }
    }
}

/* ==================================================== */
/* ==================================================== */
LOCALFUN void CreateBeats(GtkWidget *container)
{
    int i,j;
    char buffer[16];
    GtkWidget *widget;

    widget = gtk_vbox_new(FALSE,0);

    widget = gtk_table_new( MAX_DRUM_NUM,MAX_BEAT_NUM+3,0);
    GuiWidgets.DrumContainer = widget;
    
    /* gtk_container_add(GTK_CONTAINER(container),widget); */
    gtk_scrolled_window_add_with_viewport(
       GTK_SCROLLED_WINDOW(container), widget);

    for(i=0;i<MAX_DRUM_NUM;i++)
    {
        GtkWidget *parent = widget;
        GtkObject *adjustment;
        GtkWidget *widget;

        sprintf(buffer," %2d ",i);
        widget = gtk_toggle_button_new_with_label(buffer);
        gtk_object_set_data(GTK_OBJECT(widget),"trommler_drum",(gpointer) i);
        gtk_widget_set_usize(widget,30,-1);

        gtk_table_attach( GTK_TABLE(parent), widget, 0, 1, i, i+1, 0, 0, 0, 0);

        gtk_signal_connect_object( GTK_OBJECT(widget),
                                   "event",
                                   GTK_SIGNAL_FUNC(ActionDrumPopup),
                                   GTK_OBJECT(widget));
        gtk_tooltips_set_tip(GuiWidgets.Tooltips,
                             widget,
                             "Drum Popup Menu",
                             "");
        
        
        /* drum volume */
        adjustment = gtk_adjustment_new(0.0, 0.0, MAX_DRUM_VOLUME-1.0,1.0, 1.0, 1.0);
        gtk_object_set_data(GTK_OBJECT(adjustment),"trommler_drum",(gpointer) i);
        widget = gtk_spin_button_new(GTK_ADJUSTMENT(adjustment),1.0,0);
        GuiWidgets.DrumVolume[i] = widget;

        gtk_signal_connect( GTK_OBJECT(adjustment),
                                "value_changed",
                            GTK_SIGNAL_FUNC(ActionChangeDrumVolume),
                            adjustment);

        gtk_table_attach( GTK_TABLE(parent), widget, 1, 2, i, i+1, 0, 0, 0, 0);

        gtk_tooltips_set_tip(GuiWidgets.Tooltips,
                                 widget,
                                 "Set drum volume (100 is normal)",
                                 "");

        /* Panning l-r */
        adjustment = gtk_adjustment_new(0.0, -1.0, 1.0, 0.1, 1.0, 1.0);
        gtk_object_set_data(GTK_OBJECT(adjustment),"trommler_drum",(gpointer) i);
        widget = gtk_spin_button_new(GTK_ADJUSTMENT(adjustment), 0.1, 1);
        GuiWidgets.DrumPan[i] = widget;

        gtk_signal_connect( GTK_OBJECT(adjustment),
                                "value_changed",
                            GTK_SIGNAL_FUNC(ActionChangeDrumPan),
                            adjustment);

        gtk_table_attach( GTK_TABLE(parent), widget, 2, 3, i, i+1, 0, 0, 0, 0);

        gtk_tooltips_set_tip(GuiWidgets.Tooltips,
                                 widget,
                                 "Set drum panning (-1 is full left; 1 is full right)",
                                 "");

        /* mute toggle */
        widget = gtk_toggle_button_new();
        gtk_widget_set_usize(widget,15,15);
        GuiWidgets.DrumMute[i] = widget;
        gtk_table_attach( GTK_TABLE(parent), widget, 3, 4, i, i+1, 0, 0, 0, 0);
            
        gtk_object_set_data(GTK_OBJECT(widget),"trommler_drum",(gpointer) i);
        gtk_signal_connect_object( GTK_OBJECT(widget),
                                       "clicked",
                                       GTK_SIGNAL_FUNC(ActionToggleDrumMute),
                                       GTK_OBJECT(widget));

        gtk_tooltips_set_tip(GuiWidgets.Tooltips,
                                 widget,
                                 "Toggle drum mute on / off",
                                 "");
       

        /* */
        widget = gtk_entry_new();
        GuiWidgets.DrumName[i] = widget;
        gtk_object_set_data(GTK_OBJECT(widget),"trommler_drum",(gpointer) i);
        gtk_entry_set_text(GTK_ENTRY(widget),"Blechtrommel");
        gtk_entry_set_editable(GTK_ENTRY(widget),TRUE);
        
        gtk_table_attach( GTK_TABLE(parent), widget, 4, 5, i, i+1, 0, 0, 0, 0);
                
        gtk_signal_connect_object( GTK_OBJECT(widget),
                                       "activate",
                                       GTK_SIGNAL_FUNC(ActionChangeDrumName),
                                       GTK_OBJECT(widget));
        gtk_signal_connect_object( GTK_OBJECT(widget),
                                       "focus_out_event",
                                       GTK_SIGNAL_FUNC(ActionChangeDrumName),
                                       GTK_OBJECT(widget));

        for(j=0;j<MAX_BEAT_NUM;j++)
        {
#if 0
            widget = gtk_toggle_button_new_with_label(" X ");
#else
            widget = gtk_toggle_button_new();
            gtk_widget_set_usize(widget,15,15);
#endif
            gtk_object_set_data(GTK_OBJECT(widget),"trommler_drum",(gpointer) i);
            gtk_object_set_data(GTK_OBJECT(widget),"trommler_beat",(gpointer) j);
            gtk_signal_connect_object( GTK_OBJECT(widget),
                                       "clicked",
                                       GTK_SIGNAL_FUNC(ActionToggleDrumBeat),
                                       GTK_OBJECT(widget));
            
            GuiWidgets.DrumBeat[i][j] = widget;

            gtk_table_attach( GTK_TABLE(parent), widget, 5+j, 6+j, i, i+1, GTK_FILL, GTK_FILL, 0, 0);
        }
    }
}


/* ==================================================== */
/* ==================================================== */

LOCALFUN void CreateToplevel(int dimensions[4])
{
    GtkWidget *widget;
    GtkStyle *style;

    GuiWidgets.Tooltips = gtk_tooltips_new();
    gtk_tooltips_enable( GuiWidgets.Tooltips );
    
    widget = gtk_window_new(GTK_WINDOW_TOPLEVEL);
    gtk_window_set_title( GTK_WINDOW(widget), TROMMLER_INFO);

    if( dimensions[0] > 0 )
        gtk_widget_set_usize(widget,dimensions[0], dimensions[1]);

    if( dimensions[2] > 0 )
        gtk_widget_set_uposition(widget,dimensions[2], dimensions[3]);
    
    
    gtk_signal_connect_object(GTK_OBJECT(widget),"destroy", GTK_SIGNAL_FUNC(ActionQuit),NULL);

    GuiWidgets.Toplevel = widget;

    gtk_signal_connect_object(GTK_OBJECT(widget),"key_press_event",GTK_SIGNAL_FUNC(KeyboardCallback),NULL);
    gtk_widget_set_events(widget,GDK_KEY_PRESS_MASK);
       
    style = gtk_widget_get_style( widget );


    GuiWidgets.ButtonEven = gtk_style_copy(style);

    memcpy(&GuiWidgets.ButtonEven->bg[GTK_STATE_NORMAL],&style->bg[GTK_STATE_ACTIVE],sizeof(GdkColor) );
    memcpy(&GuiWidgets.ButtonEven->bg[GTK_STATE_PRELIGHT],&style->bg[GTK_STATE_ACTIVE],sizeof(GdkColor) );
    memcpy(&GuiWidgets.ButtonEven->bg[GTK_STATE_ACTIVE],&style->black,sizeof(GdkColor) );
    GuiWidgets.ButtonEven->bg_gc[GTK_STATE_NORMAL] = style->bg_gc[GTK_STATE_ACTIVE];
    GuiWidgets.ButtonEven->bg_gc[GTK_STATE_PRELIGHT] = style->bg_gc[GTK_STATE_ACTIVE];
    GuiWidgets.ButtonEven->bg_gc[GTK_STATE_ACTIVE] = style->black_gc;

    GuiWidgets.ButtonOdd = gtk_style_copy(style);

    memcpy(&GuiWidgets.ButtonOdd->bg[GTK_STATE_NORMAL],&style->bg[GTK_STATE_PRELIGHT],sizeof(GdkColor) );
    memcpy(&GuiWidgets.ButtonOdd->bg[GTK_STATE_PRELIGHT],&style->bg[GTK_STATE_PRELIGHT],sizeof(GdkColor) );
    memcpy(&GuiWidgets.ButtonOdd->bg[GTK_STATE_ACTIVE],&style->black,sizeof(GdkColor) );
    GuiWidgets.ButtonOdd->bg_gc[GTK_STATE_NORMAL] = style->bg_gc[GTK_STATE_PRELIGHT];
    GuiWidgets.ButtonOdd->bg_gc[GTK_STATE_PRELIGHT] = style->bg_gc[GTK_STATE_PRELIGHT];
    GuiWidgets.ButtonOdd->bg_gc[GTK_STATE_ACTIVE] = style->black_gc;
        
    /* Vbox within toplevel */
    {
        GtkWidget *parent = widget;
        GtkWidget *widget = gtk_vbox_new(FALSE,0);
        gtk_container_add( GTK_CONTAINER(parent), widget);
        
        /* Menubar */
        {
            GtkWidget *parent = widget;
            GtkWidget *widget = gtk_hbox_new(FALSE,0);
            gtk_box_pack_start( GTK_BOX(parent), widget, FALSE,FALSE,0);
            
            /* Menubar Items */
            {
                GtkWidget *parent = widget;
                GtkWidget *widget;

                widget = gtk_button_new_with_label(" Quit ");
                gtk_box_pack_start( GTK_BOX(parent), widget,FALSE,FALSE,0);
                gtk_signal_connect_object(GTK_OBJECT(widget),
                                          "clicked",
                                          GTK_SIGNAL_FUNC(ActionQuit),
                                          NULL);
                

                widget = gtk_button_new_with_label(" About ... ");
                gtk_box_pack_start( GTK_BOX(parent), widget,FALSE,FALSE,0);
                gtk_signal_connect_object(GTK_OBJECT(widget),
                                          "clicked",
                                          GTK_SIGNAL_FUNC(ActionAbout),
                                          NULL);

                widget = gtk_label_new("   ");
                gtk_box_pack_start( GTK_BOX(parent), widget, FALSE,FALSE,0);

                
                widget = gtk_button_new_with_label(" New Song ");
                gtk_box_pack_start( GTK_BOX(parent), widget,FALSE,FALSE,0);
                gtk_signal_connect_object(GTK_OBJECT(widget),
                                          "clicked",
                                          GTK_SIGNAL_FUNC(ActionNewSong),
                                          NULL);
                
                widget = gtk_button_new_with_label(" Load Song ... ");
                gtk_box_pack_start( GTK_BOX(parent), widget,FALSE,FALSE,0);
                gtk_signal_connect_object(GTK_OBJECT(widget),
                                          "clicked",
                                          GTK_SIGNAL_FUNC(ActionLoadSong),
                                          NULL);
                
                widget = gtk_button_new_with_label(" Save Song As ... ");
                gtk_box_pack_start( GTK_BOX(parent), widget,FALSE,FALSE,0);
                gtk_signal_connect_object(GTK_OBJECT(widget),
                                          "clicked",
                                          GTK_SIGNAL_FUNC(ActionSaveSongAs),
                                          NULL);

                widget = gtk_button_new_with_label(" Save Song ");
                gtk_box_pack_start( GTK_BOX(parent), widget,FALSE,FALSE,0);
                gtk_signal_connect_object(GTK_OBJECT(widget),
                                          "clicked",
                                          GTK_SIGNAL_FUNC(ActionSaveSong),
                                          NULL);

                widget = gtk_label_new("   Save Sample   ");
                gtk_box_pack_start( GTK_BOX(parent), widget, FALSE,FALSE,0);
                
                widget = gtk_button_new_with_label(" Pattern ... ");
                gtk_box_pack_start( GTK_BOX(parent), widget,FALSE,FALSE,0);
                gtk_signal_connect_object( GTK_OBJECT(widget),
                                           "clicked",
                                           GTK_SIGNAL_FUNC(ActionSaveSamplePattern),
                                           NULL);

                widget = gtk_button_new_with_label(" Range ... ");
                gtk_box_pack_start( GTK_BOX(parent), widget,FALSE,FALSE,0);
                gtk_signal_connect_object( GTK_OBJECT(widget),
                                           "clicked",
                                           GTK_SIGNAL_FUNC(ActionSaveSampleRange),
                                           NULL);
            
                widget = gtk_button_new_with_label(" Song ... ");
                gtk_box_pack_start( GTK_BOX(parent), widget,FALSE,FALSE,0);
                gtk_signal_connect_object( GTK_OBJECT(widget),
                                           "clicked",
                                           GTK_SIGNAL_FUNC(ActionSaveSampleSong),
                                           NULL);

            }
        }
        
        /* Pattern Menu */
        {
            GtkObject *adjustment;
            GtkWidget *parent = widget;
            GtkWidget *widget = gtk_hbox_new(FALSE,0);
            gtk_box_pack_start( GTK_BOX(parent), widget,FALSE,FALSE,0);
            
            /* Pattern Menu Items */
            {
                GtkWidget *parent = widget;
                GtkWidget *widget;

                widget = gtk_label_new("Pattern:");
                gtk_box_pack_start( GTK_BOX(parent), widget, FALSE,FALSE,0);

                adjustment = gtk_adjustment_new(0.0, 0.0, MAX_PATT_NUM-1.0, 1.0, 1.0, 0.0);
                widget = gtk_spin_button_new(GTK_ADJUSTMENT(adjustment),1.0,0);
                GuiWidgets.PatternNumber = widget;

                gtk_signal_connect( GTK_OBJECT(adjustment),
                                    "value_changed",
                                    GTK_SIGNAL_FUNC(ActionChangePatternNumber),
                                    adjustment);

                gtk_box_pack_start( GTK_BOX(parent), widget, FALSE,FALSE,0);
                gtk_tooltips_set_tip(GuiWidgets.Tooltips,
                                     widget,
                                     "Choose current pattern",
                                     "");
                
                
                /* pattern name change entry */
                
                widget = gtk_entry_new();
                GuiWidgets.PatternName = widget;
                gtk_entry_set_text(GTK_ENTRY(widget),"0123456789012345");
                gtk_entry_set_editable(GTK_ENTRY(widget),TRUE);
                gtk_entry_set_max_length(GTK_ENTRY(widget),16);
                gtk_signal_connect_object( GTK_OBJECT(widget),
                                           "activate",
                                           GTK_SIGNAL_FUNC(ActionChangePatternName),
                                           GTK_OBJECT(widget));
                gtk_signal_connect_object( GTK_OBJECT(widget),
                                           "focus_out_event",
                                           GTK_SIGNAL_FUNC(ActionChangePatternName),
                                           GTK_OBJECT(widget));
                gtk_box_pack_start( GTK_BOX(parent), widget, FALSE,FALSE,0);
                gtk_tooltips_set_tip(GuiWidgets.Tooltips,
                                     widget,
                                     "Change pattern description",
                                     "");

                                /* pattern length  */
                
                widget = gtk_label_new("  Units:");
                gtk_box_pack_start( GTK_BOX(parent), widget, FALSE,FALSE,0);

                adjustment = gtk_adjustment_new(1.0, 1.0, MAX_UNITS_NUM, 1.0, 1.0, 0.0);
                widget = gtk_spin_button_new(GTK_ADJUSTMENT(adjustment),1.0,0);

                GuiWidgets.PatternUnits = widget;

                gtk_signal_connect( GTK_OBJECT(adjustment),
                                    "value_changed",
                                    GTK_SIGNAL_FUNC(ActionChangePatternUnits),
                                    adjustment);


                                
                gtk_box_pack_start( GTK_BOX(parent), widget, FALSE,FALSE,0);


                gtk_tooltips_set_tip(GuiWidgets.Tooltips,
                                     widget,
                                     "Set number of units in current pattern",
                                     "");

                /* pattern length  */
                
                widget = gtk_label_new("  BPUs:");
                gtk_box_pack_start( GTK_BOX(parent), widget, FALSE,FALSE,0);

                adjustment = gtk_adjustment_new(1.0, 1.0, MAX_BEATS_PER_UNIT_NUM, 1.0, 1.0, 0.0);
                widget = gtk_spin_button_new(GTK_ADJUSTMENT(adjustment),1.0,0);

                GuiWidgets.PatternBeatsPerUnit = widget;

                gtk_signal_connect( GTK_OBJECT(adjustment),
                                    "value_changed",
                                    GTK_SIGNAL_FUNC(ActionChangePatternBeatsPerUnit),
                                    adjustment);

                gtk_box_pack_start( GTK_BOX(parent), widget, FALSE,FALSE,0);


                gtk_tooltips_set_tip(GuiWidgets.Tooltips,
                                     widget,
                                     "Set number of beats per unit in current pattern",
                                     "");

                /* filler */
                
                widget = gtk_label_new("   ");
                gtk_box_pack_start( GTK_BOX(parent), widget, FALSE,FALSE,0);

                /* pattern rotation left */
                
                widget = gtk_button_new_with_label(" < ");
                gtk_box_pack_start( GTK_BOX(parent), widget, FALSE,FALSE,0);
                gtk_signal_connect_object( GTK_OBJECT(widget),
                                           "clicked",
                                           GTK_SIGNAL_FUNC(ActionBeatsRotateLeft),
                                           GTK_OBJECT(widget));
                gtk_tooltips_set_tip(GuiWidgets.Tooltips,
                                     widget,
                                     "Rotate beats of current pattern to the left",
                                     "");

                widget = gtk_label_new("Rotate");
                gtk_box_pack_start( GTK_BOX(parent), widget, FALSE,FALSE,0);

                /* pattern rotation right */
                
                widget = gtk_button_new_with_label(" > ");
                gtk_box_pack_start( GTK_BOX(parent), widget, FALSE,FALSE,0);
                gtk_signal_connect_object( GTK_OBJECT(widget),
                                           "clicked",
                                           GTK_SIGNAL_FUNC(ActionBeatsRotateRight),
                                           GTK_OBJECT(widget));
                gtk_tooltips_set_tip(GuiWidgets.Tooltips,
                                      widget,
                                      "Rotate beats of current pattern to the right",
                                      "");

                /* filler */
                
                widget = gtk_label_new("   ");
                gtk_box_pack_start( GTK_BOX(parent), widget, FALSE,FALSE,0);
                
                /* copy pattern */
                
                widget = gtk_button_new_with_label(" Copy ");
                gtk_box_pack_start( GTK_BOX(parent), widget, FALSE,FALSE,0);
                gtk_signal_connect_object( GTK_OBJECT(widget),
                                           "clicked",
                                           GTK_SIGNAL_FUNC(ActionCopyPattern),
                                           GTK_OBJECT(widget));
                gtk_tooltips_set_tip(GuiWidgets.Tooltips,
                                     widget,
                                     "Copy other pattern into current pattern",
                                     "");


                /* choose source pattern */

                adjustment = gtk_adjustment_new(0.0, 0.0, MAX_PATT_NUM-1.0, 1.0, 1.0, 0.0);
                widget = gtk_spin_button_new(GTK_ADJUSTMENT(adjustment),1.0,0);
                GuiWidgets.PatternSource = widget;
                gtk_box_pack_start( GTK_BOX(parent), widget, FALSE,FALSE,0);

                gtk_tooltips_set_tip(GuiWidgets.Tooltips,
                                     widget,
                                     "Choose source pattern for copy",
                                     "");

                gtk_signal_connect( GTK_OBJECT(adjustment),
                                    "value_changed",
                                    GTK_SIGNAL_FUNC(ActionChangeCopyPatternNumber),
                                    adjustment);

                /* pattern name change entry */
                
                widget = gtk_entry_new();
                GuiWidgets.CopyPatternName = widget;
                gtk_entry_set_text(GTK_ENTRY(widget),"*Copy Source*");
                gtk_entry_set_editable(GTK_ENTRY(widget),FALSE);
                gtk_entry_set_max_length(GTK_ENTRY(widget),16);
                gtk_box_pack_start( GTK_BOX(parent), widget, FALSE,FALSE,0);
                
            }
        }
        
        /* Beats */
        {
            GtkWidget *parent = widget;
            GtkWidget *widget = gtk_scrolled_window_new(NULL,NULL);
            gtk_box_pack_start( GTK_BOX(parent), widget, TRUE,TRUE,0);
            gtk_scrolled_window_set_policy( GTK_SCROLLED_WINDOW(widget),
                                            GTK_POLICY_AUTOMATIC,
                                            GTK_POLICY_AUTOMATIC);

            CreateBeats(widget);
        }

        /* Song Menu */
        {
            GtkObject *adjustment;
            GtkWidget *parent = widget;
            GtkWidget *widget = gtk_hbox_new(FALSE,0);
            gtk_box_pack_start( GTK_BOX(parent), widget, FALSE,FALSE,0);
            
            /* song Menu Items */
            {
                GtkWidget *parent = widget;
                GtkWidget *widget;

                widget = gtk_label_new("Song:");
                gtk_box_pack_start( GTK_BOX(parent), widget, FALSE,FALSE,0);
                
                widget = gtk_entry_new();
                GuiWidgets.SongName = widget;
                gtk_entry_set_text(GTK_ENTRY(widget),"The song");
                gtk_entry_set_editable(GTK_ENTRY(widget),TRUE);
                gtk_entry_set_max_length(GTK_ENTRY(widget),16);
                gtk_signal_connect_object( GTK_OBJECT(widget),
                                           "activate",
                                           GTK_SIGNAL_FUNC(ActionChangeSongName),
                                           NULL);
                gtk_signal_connect_object( GTK_OBJECT(widget),
                                           "focus_out_event",
                                           GTK_SIGNAL_FUNC(ActionChangeSongName),
                                           NULL);
                gtk_box_pack_start( GTK_BOX(parent), widget, FALSE,FALSE,0);
                gtk_tooltips_set_tip(GuiWidgets.Tooltips,
                                     widget,
                                     "change song description",
                                     "");
                /* */
                
                widget = gtk_label_new("  Measures:");
                gtk_box_pack_start( GTK_BOX(parent), widget, FALSE,FALSE,0);

                /* */
                
                adjustment = gtk_adjustment_new(0.0, 0.0, MAX_MEASURE_NUM, 1.0, 1.0, 1.0);
                widget = gtk_spin_button_new(GTK_ADJUSTMENT(adjustment),1.0,0);
                GuiWidgets.SongMeasures = widget;

                gtk_signal_connect_object( GTK_OBJECT(adjustment),
                                           "value_changed",
                                           GTK_SIGNAL_FUNC(ActionChangeSongLength),
                                           GTK_OBJECT(adjustment));

                gtk_box_pack_start( GTK_BOX(parent), widget, FALSE,FALSE,0);
                gtk_tooltips_set_tip(GuiWidgets.Tooltips,
                                     widget,
                                     "Set number of measure in song",
                                     "");

                /* */
                
                widget = gtk_label_new("  UPMs:");
                gtk_box_pack_start( GTK_BOX(parent), widget, FALSE,FALSE,0);

                /* */
                
                adjustment = gtk_adjustment_new (1.0, 1.0, MAX_UPM_NUM-1.0, 1.0, 1.0, 0.0);
                widget = gtk_spin_button_new(GTK_ADJUSTMENT(adjustment),1.0,0);
                GuiWidgets.SongBpms = widget;

                gtk_signal_connect_object( GTK_OBJECT(adjustment),
                                           "value_changed",
                                           GTK_SIGNAL_FUNC(ActionChangeSongUPMS),
                                           GTK_OBJECT(adjustment));


                gtk_box_pack_start( GTK_BOX(parent), widget, FALSE,FALSE,0);
                gtk_tooltips_set_tip(GuiWidgets.Tooltips,
                                     widget,
                                     "Set beats per minute in song",
                                     "");

                /* */
                
                widget = gtk_label_new(" Range: ");
                gtk_box_pack_start( GTK_BOX(parent), widget, FALSE,FALSE,0);

                /* */
                
                adjustment = gtk_adjustment_new (0.0, 0.0, MAX_MEASURE_NUM-1.0, 1.0, 1.0, 0.0);
                widget = gtk_spin_button_new(GTK_ADJUSTMENT(adjustment),1.0,0);
                GuiWidgets.SongRangeStart = widget;

                gtk_signal_connect_object( GTK_OBJECT(adjustment),
                                           "value_changed",
                                           GTK_SIGNAL_FUNC(ActionChangeSongRangeStart),
                                           GTK_OBJECT(adjustment));


                gtk_box_pack_start( GTK_BOX(parent), widget, FALSE,FALSE,0);
                gtk_tooltips_set_tip(GuiWidgets.Tooltips,
                                     widget,
                                     "Set first measure of song range",
                                     "");

                /* */
                
                widget = gtk_label_new("  to ");
                gtk_box_pack_start( GTK_BOX(parent), widget, FALSE,FALSE,0);

                /* */
                
                adjustment = gtk_adjustment_new (0.0, 0.0, MAX_MEASURE_NUM-1.0, 1.0, 1.0, 0.0);
                widget = gtk_spin_button_new(GTK_ADJUSTMENT(adjustment),1.0,0);
                GuiWidgets.SongRangeStop = widget;

                gtk_signal_connect_object( GTK_OBJECT(adjustment),
                                           "value_changed",
                                           GTK_SIGNAL_FUNC(ActionChangeSongRangeStop),
                                           GTK_OBJECT(adjustment));


                gtk_box_pack_start( GTK_BOX(parent), widget, FALSE,FALSE,0);
                gtk_tooltips_set_tip(GuiWidgets.Tooltips,
                                     widget,
                                     "Set last measure of song range",
                                     "");
                /* filler */
                
                widget = gtk_label_new("   ");
                gtk_box_pack_end( GTK_BOX(parent), widget, FALSE,FALSE,0);

                /* song path*/
                
                widget = gtk_label_new("a_song_file.sng");
                GuiWidgets.SongFile = widget;
                gtk_box_pack_end( GTK_BOX(parent), widget,FALSE,FALSE,0);
                gtk_widget_show(widget);


            }
            
            gtk_widget_show(widget);
        }
        
        /* Song */
        {
            GtkWidget *parent = widget;
            GtkWidget *widget = gtk_scrolled_window_new(NULL,NULL);
            gtk_box_pack_start( GTK_BOX(parent), widget, FALSE,FALSE,0);
            gtk_scrolled_window_set_policy( GTK_SCROLLED_WINDOW(widget),
                                            GTK_POLICY_ALWAYS,
                                            GTK_POLICY_AUTOMATIC);
            gtk_widget_set_usize(widget,-1,80);
            CreateMeasures(widget);
        }
        
        /* Play Menu */
        {
            GtkWidget *parent = widget;
            GtkWidget *widget = gtk_hbox_new(FALSE,0);
            gtk_box_pack_start( GTK_BOX(parent), widget, FALSE,FALSE,0);
            
            /* Play Menu Items */
            {
                GtkWidget *parent = widget;
                GtkWidget *widget;
                GtkWidget *leader;
                
                widget = gtk_radio_button_new_with_label( gtk_radio_button_group(NULL),"Pattern");
                leader =  widget;

                gtk_box_pack_start(GTK_BOX(parent),widget,FALSE,FALSE,0);
                gtk_signal_connect_object( GTK_OBJECT(widget),
                                           "clicked",
                                           GTK_SIGNAL_FUNC(ActionPlayPattern),
                                           NULL);

                ActionPlayPattern();
                
                widget = gtk_radio_button_new_with_label(gtk_radio_button_group(GTK_RADIO_BUTTON (leader)),"Range");
                gtk_box_pack_start(GTK_BOX(parent),widget,FALSE,FALSE,0);
                gtk_signal_connect_object( GTK_OBJECT(widget),
                                           "clicked",
                                           GTK_SIGNAL_FUNC(ActionPlayRange),
                                           NULL);


                widget = gtk_radio_button_new_with_label(gtk_radio_button_group(GTK_RADIO_BUTTON (leader)),"Song");
                gtk_box_pack_start(GTK_BOX(parent),widget,FALSE,FALSE,0);
                gtk_signal_connect_object( GTK_OBJECT(widget),
                                           "clicked",
                                           GTK_SIGNAL_FUNC(ActionPlaySong),
                                           NULL);

                /* filler */
                
                widget = gtk_label_new("   ");
                gtk_box_pack_end( GTK_BOX(parent), widget, FALSE,FALSE,0);
                
                /* Play Position */
                
                widget = gtk_label_new("000:000:000");
                GuiWidgets.PlayDisplay = widget;
                gtk_box_pack_end( GTK_BOX(parent), widget,FALSE,FALSE,0);

                /* filler */
                
                widget = gtk_label_new("   ");
                gtk_box_pack_end( GTK_BOX(parent), widget, FALSE,FALSE,0);

                /* Reset */

                widget = gtk_button_new_with_label(" Reset ");
                gtk_box_pack_end( GTK_BOX(parent), widget,FALSE,FALSE,0);
                gtk_signal_connect_object(GTK_OBJECT(widget),
                                          "clicked",
                                          GTK_SIGNAL_FUNC(ActionPlayReset),
                                          NULL);

                /* filler */
                
                widget = gtk_label_new("   ");
                gtk_box_pack_end( GTK_BOX(parent), widget, FALSE,FALSE,0);

                /* Loop */

                widget = gtk_toggle_button_new_with_label(" Loop ");
                GuiWidgets.PlayLoopButton = widget;
                gtk_box_pack_end( GTK_BOX(parent), widget,FALSE,FALSE,0);
                gtk_signal_connect_object(GTK_OBJECT(widget),
                                          "toggled",
                                          GTK_SIGNAL_FUNC(ActionPlayLoop),
                                          NULL);

                ActionPlayLoop(); // set state to non active
                

                /* Active */

                widget = gtk_toggle_button_new_with_label(" Start ");
                GuiWidgets.PlayActiveButton = widget;
                gtk_box_pack_end( GTK_BOX(parent), widget,FALSE,FALSE,0);
                gtk_signal_connect_object(GTK_OBJECT(widget),
                                          "toggled",
                                          GTK_SIGNAL_FUNC(ActionPlayActive),
                                          NULL);

                ActionPlayActive(); // set state to non active
            }
        }
    }
}    

/* ==================================================== */
LOCALFUN void StopPlaying()
{
    ActionPlayReset();
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(GuiWidgets.PlayActiveButton), 0);
}

/* ==================================================== */
#define B_OK 0
#define B_FAIL 1

LOCALFUN INT32 HandleBeatIncrements()
{
    switch( GuiCurrent.PlayMode )
    {
      case PLAY_PATTERN:
        GuiCurrent.PlayMeasure = 0;
        GuiCurrent.PlayPattern = GuiCurrent.Pattern;

        if( GuiCurrent.PlayBeat >= PATT_BEATS(GuiCurrent.PlayPattern) )
        {
            if( GuiCurrent.PlayLoop )
            {
                GuiCurrent.PlayBeat = 0;
            }
            else
            {
                return B_FAIL;
            }
            
        }
        
        return B_OK;

      case PLAY_RANGE:
        if( GuiCurrent.StartMeasure >= SONG_LENGTH ||
            GuiCurrent.StartMeasure < 0 ||
            GuiCurrent.StopMeasure >=  SONG_LENGTH ||
            GuiCurrent.StopMeasure < 0 ||
            GuiCurrent.StartMeasure > GuiCurrent.StopMeasure )
        {
            return B_FAIL;
        }
        
        if( GuiCurrent.PlayMeasure > GuiCurrent.StopMeasure ||
            GuiCurrent.PlayMeasure < GuiCurrent.StartMeasure )
            GuiCurrent.PlayMeasure = GuiCurrent.StartMeasure;

        GuiCurrent.PlayPattern = SONG_MEASURE[GuiCurrent.PlayMeasure];

        if( GuiCurrent.PlayBeat >= PATT_BEATS(GuiCurrent.PlayPattern) )
        {
            GuiCurrent.PlayBeat = 0;
            GuiCurrent.PlayMeasure++;
            if( GuiCurrent.PlayMeasure > GuiCurrent.StopMeasure )
            {
                if (GuiCurrent.PlayLoop)
                {
                   GuiCurrent.PlayMeasure = GuiCurrent.StartMeasure;
                   GuiCurrent.PlayPattern = SONG_MEASURE[GuiCurrent.PlayMeasure];
                }
                else
                {
                    return B_FAIL;
                }
            }            
        }
        return B_OK;

      case PLAY_SONG:
        if( GuiCurrent.PlayMeasure >= SONG_LENGTH )
            GuiCurrent.PlayMeasure = 0;
        
        GuiCurrent.PlayPattern = SONG_MEASURE[GuiCurrent.PlayMeasure];
        if( GuiCurrent.PlayBeat >= PATT_BEATS(GuiCurrent.PlayPattern) )
        {
            GuiCurrent.PlayBeat = 0;
            GuiCurrent.PlayMeasure++;
            if( GuiCurrent.PlayMeasure >= SONG_LENGTH )
            {
                if (GuiCurrent.PlayLoop)
                {
                    GuiCurrent.PlayMeasure = 0;
                    GuiCurrent.PlayPattern = SONG_MEASURE[GuiCurrent.PlayMeasure];
                }
                else
                {
                    return B_FAIL;
                }
            }            
        }
        return B_OK;

      default:
        ASSERTX("unexpected case");
        return B_FAIL;
    }
}

/* ==================================================== */

LOCALFUN gint SoundTimerInterrupt()
{
    if( GuiCurrent.PlayActive )
    {
        const int BufferBytesTotal = SoundcardBufferTotal();

        int BufferBytesAvail = SoundcardBufferAvail();

        do
        {
            int BytesCurrent;
            
            if( B_FAIL == HandleBeatIncrements() )
            {
                StopPlaying();
                break;
            }

            /* compute the sample size for the current beat */
               
            BytesCurrent = 44100*2*60 / (SONG_UPMS * PATT_BEATS_PER_UNIT(GuiCurrent.PlayPattern));

            // if no space delay playing this beat
            if( BytesCurrent >= BufferBytesAvail && BufferBytesAvail <= BufferBytesTotal/2)  break;
            
            BufferBytesAvail -= BytesCurrent;
                
            SetPlay();

            PlayBeat(NULL,
                     GuiCurrent.PlayPattern,
                     GuiCurrent.PlayBeat,
                     SONG_UPMS*PATT_BEATS_PER_UNIT(GuiCurrent.PlayPattern));
            
            GuiCurrent.PlayBeat++; /* this increment is handled by  HandleBeatIncrements() */
        }
        while( BufferBytesAvail > 0 );

    }

    gtk_timeout_add(TIMER_MS_INTERVAL,SoundTimerInterrupt,NULL);
    return FALSE;
}

/* ==================================================== */
/* ==================================================== */

LOCALFUN void parse_geometry(char *string, int dimensions[4])
{
    char *pos = string;
    int i;
#if 0
    GdkScreen*  screen = gdk_get_default_screen();
#endif
    for( i=0; i<4; i++ )
    {
        if( *pos == 'x' ) pos++;
        if( *pos == 'X' ) pos++;
        if( *pos == '\0' ) break;
        
        dimensions[i] = strtol(pos, &pos, 10);
#if 0
        if( i == 2 && dimensions[i] < 0 )
            dimensions[i] += gdk_screen_get_width(screen);
        if( i == 3 && dimensions[i] < 0 )
            dimensions[i] += gdk_screen_get_height(screen);
#endif
    }
}

/* ==================================================== */
/* ==================================================== */
GLOBALFUN int main(int argc, char *argv[])
{
    /* dimension and position of main window
       may be overwritten with geometry option */
    
    int dimensions[4] = { 800, 600, -1,  -1 };
    char *song =  NULL;
    int i;
    
    gtk_init(&argc, &argv);

    for(i=1; i<argc; i++ )
    {
        if( 0==strcmp(argv[i],"-geometry") )
        {
            i++;
            ASSERT( i < argc,"bad geometry arg");
            parse_geometry(argv[i],dimensions);
        }
        else
        {
            song = argv[i];
        }
    }
    
    printf("++ Create Widgets %d %d %d %d\n",dimensions[0],dimensions[1],dimensions[2],dimensions[3]);
    CreateToplevel(dimensions);
    printf("-- Done\n");

    printf("++ Show Widgets\n");
    gtk_widget_show_all(GuiWidgets.Toplevel);
    printf("-- Done\n");

    
    printf("++ Prefix Drum Path\n");

    if( NULL != getenv("TROMMLER_DRUM_DIR") )
    {
        strcpy(Global.DrumDir,getenv("TROMMLER_DRUM_DIR"));
    }
    else
    {
        getcwd(Global.DrumDir,sizeof(Global.DrumDir));
        strncat(Global.DrumDir,"/Drums",sizeof(Global.DrumDir) );
    }
    
    printf("-- Done %s\n",Global.DrumDir);

    memset(Global.SampleDir, 0, sizeof(Global.SampleDir));
    
    if( song )
    {
        SongLoad(song);
        SetSong();
    }
    else
    {
        ActionNewSong();
    }

    printf("++ Initialize Soundcard\n");
    SoundcardOpen();
    printf("-- Done\n");

    SoundTimerInterrupt();
    gtk_main();

    return 0;
}


/* ==================================================== */
/* ==================================================== */
